@@ -11,23 +11,26 @@
 #include "aesni.h"
 #include "secmem.h"
 #include <string.h>
-#include <wmmintrin.h>
+
 #include "archdep.h"
 
-
-#ifdef __AVX2__
+#if defined(__x86_64__) || defined(__i386__) || defined(__i486__) || defined(__i586__) || defined(__i686__)
+#include <wmmintrin.h>
+#if defined(__AVX2__) && !defined(NO_AVX2)
 #include <immintrin.h>
 static int probe_aes_ni()
 {
 	return !have_aesni || !have_avx2;
 }
-#else
+#else	/* AVX2 */
 #include <smmintrin.h>
 static int probe_aes_ni()
 {
 	return !have_aesni;
 }
-#endif
+#endif	/* AVX2 */
+#endif	/* x86 */
+
 /* Unaligned 128bit integer type, missing from gcc < 5 emmintrin.h */
 typedef long long __m128i_u __attribute__ ((__vector_size__ (16), __may_alias__, __aligned__ (1)));
 
@@ -85,7 +88,7 @@ typedef long long __m128i_u __attribute__ ((__vector_size__ (16), __may_alias__,
 #define MMCLEARALL_MAN2 do {} while(0)
 #endif
 
-#ifdef __AVX2__
+#if defined(__AVX2__) && !defined(NO_AVX2)
 #define MMCLEARALL _mm256_zeroall()
 #define MM256CLEAR5						\
 	asm volatile("	vpxor %%ymm4, %%ymm4, %%ymm4	\n"	\
@@ -717,7 +720,7 @@ int  AESNI_ECB_Crypt_Tmpl(crypt_8blks_fn *crypt8, crypt_blk_fn *crypt, int enc,
 		return 0;
 }
 
-#if defined(__AVX2__) && defined(__VAES__)
+#if defined(__AVX2__) && !defined(NO_AVX2) && defined(__VAES__)
 #ifdef __GNUC__
 #define ALWAYS_INLINE __attribute__ ((__gnu_inline__, __always_inline__, __artificial__))
 #else
@@ -1124,8 +1127,8 @@ int AESNI_CTR_Crypt_Tmpl(crypt_8blks_fn *crypt8, crypt_blk_fn *crypt,
 #ifdef AESNI_PREFETCH
 		__builtin_prefetch(outptr, 1, 3);
 		__builtin_prefetch(outptr+4, 1, 3);
-		__builtin_prefetch(inptr, 0, 1);
-		__builtin_prefetch(inptr+4, 0, 1);
+		//__builtin_prefetch(inptr, 0, 1);
+		//__builtin_prefetch(inptr+4, 0, 1);
 #endif
 		/* Prepare CTR (IV) values */
 		__m128i tmp0 = _mm_shuffle_epi8(cblk, BSWAP_EPI64);
@@ -1182,7 +1185,7 @@ int AESNI_CTR_Crypt_Tmpl(crypt_8blks_fn *crypt8, crypt_blk_fn *crypt,
 	return 0;
 }
 
-#if !defined(__AVX2__) || !defined(__VAES__)
+#if !defined(__AVX2__) || !defined(__VAES__) || defined(NO_AVX2)
 static
 int AESNI_CTR_Crypt(const unsigned char* key, unsigned int rounds,
 		     unsigned char* ctr, unsigned int pad,
@@ -1220,8 +1223,8 @@ int AESNI_CTR_Crypt_Tmpl2(crypt_4x2blks_fn *crypt4, crypt_blk_fn *crypt,
 #ifdef AESNI_PREFETCH
 		__builtin_prefetch(outptr, 1, 3);
 		__builtin_prefetch(outptr+2, 1, 3);
-		__builtin_prefetch(inptr, 0, 1);
-		__builtin_prefetch(inptr+2, 0, 1);
+		//__builtin_prefetch(inptr, 0, 1);
+		//__builtin_prefetch(inptr+2, 0, 1);
 #endif
 		/* Prepare CTR (IV) values */
 		tmp0 = _mm256_shuffle_epi8(cblk, BSWAP_BOTH);
@@ -1690,7 +1693,7 @@ int  AESNI_CTR_CryptX2(const uchar* rkeys, unsigned int rounds,
 				    rkeys, rounds, iv, in, out, len);
 }
 
-#ifdef __AVX2__
+#if defined(__AVX2__) && !defined(NO_AVX2)
 #define AESNI_METHODS VAESNI_Methods
 #else
 #define AESNI_METHODS SAESNI_Methods
