@@ -15,11 +15,11 @@
 
 #include "ddr_plugin.h"
 #include "ddr_ctrl.h"
+#include "mybasename.h"
 
 #include <stdlib.h>
 #include <string.h>
 #include <stdint.h>
-#include <libgen.h>
 #include <unistd.h>
 #include <string.h>
 #include <assert.h>
@@ -27,12 +27,14 @@
 #include <netinet/in.h>
 #include <sys/stat.h>
 #include <signal.h>
+#ifdef HAVE_LZO_LZO1X_H	// avoid breaking on dependency generation
 #include <lzo/lzo1x.h>
 #include <lzo/lzo1y.h>
 #include <lzo/lzo1f.h>
 #include <lzo/lzo1b.h>
 #include <lzo/lzo1c.h>
 #include <lzo/lzo2a.h>
+#endif
 #include <time.h>
 #ifdef HAVE_ENDIAN_H
 #include <endian.h>
@@ -46,25 +48,9 @@
 #error __WORDSIZE unknown
 #endif
 
-#ifdef HAVE_BASENAME
-//char* basename(char*);
-#else
-static char* basename(char *nm)
-{
-	const char* ptr = strrchr(nm, '/');	/* Not on DOS */
-	if (ptr)
-		return ptr+1;
-	else
-		return nm;
-}
-#endif
-
 // TODO: pass at runtime rather than compile time
-#ifdef DEBUG
-# define LZO_DEBUG(x) x
-#else
-# define LZO_DEBUG(x) do {} while (0)
-#endif
+#define LZO_DEBUG(x) \
+	do { if (state->debug) x; } while(0)
 
 /* Some bits from lzop -- we strive for some level of compatibility */
 /* We use version numbers that are not likely to clash with lzop anytime soon;
@@ -115,11 +101,11 @@ typedef struct
 } __attribute__((packed)) header_t;
 
 typedef struct {
-    uint32_t uncmpr_len;   
-    uint32_t cmpr_len;   
+    uint32_t uncmpr_len;
+    uint32_t cmpr_len;
     uint32_t uncmpr_chksum;
     uint32_t cmpr_chksum;
-} blockhdr_t;
+} __attribute__((packed)) blockhdr_t;
 
 #define ADLER32_INIT_VALUE 1
 #define CRC32_INIT_VALUE 0
@@ -190,7 +176,7 @@ enum compmode {AUTO=0, COMPRESS, DECOMPRESS};
 
 typedef struct _lzo_state {
 	void *workspace;
-	unsigned char *dbuf, *orig_dbuf;
+	unsigned char *dbuf;
 	unsigned char *obuf;
 	size_t dbuflen;
        	int hdroff;
@@ -201,6 +187,7 @@ typedef struct _lzo_state {
 	unsigned int blockno, holeno;
 	unsigned char eof_seen, do_bench, do_opt, do_search;
 	unsigned char debug, nodiscard;
+	unsigned char islast;
 	enum compmode mode;
 	unsigned int last_ulen;
 	comp_alg *algo;
@@ -212,10 +199,12 @@ typedef struct _lzo_state {
 	size_t cmp_ln, unc_ln;
 	/* Bench */
 	clock_t cpu;
+	/* Unsparse */
+	loff_t inhole;
+	unsigned char *buf_zero;
+	int saved_c_off;
 } lzo_state;
 
-#define FPLOG(lvl, fmt, args...) \
-	plug_log(ddr_plug.logger, stderr, lvl, fmt, ##args)
 
 static unsigned int pagesize = 4096;
 
@@ -233,7 +222,7 @@ void lzo_hdr(header_t* hdr, loff_t hole, lzo_state *state)
 	hdr->flags = htonl(state->flags);
 	hdr->nmlen = NAMELEN;
 	if (hole) {
-		char* bnm = basename((char*)state->opts->iname);
+		const char* bnm = mybasename(state->opts->iname);
 		/* This would abort with -D_FORTIFY_SOURCE=2 
 		sprintf(hdr->name+6, ".%04x.%010lx", state->holeno++, hole);
 		*/
@@ -246,9 +235,9 @@ void lzo_hdr(header_t* hdr, loff_t hole, lzo_state *state)
 		hdr->mtime_low = htonl(hole & 0xffffffff);
 		hdr->mtime_high= htonl(hole >> 32);
 	} else {
-		char* nm = (char*)state->opts->iname;
+		const char* nm = (char*)state->opts->iname;
 		if (strlen(nm) > NAMELEN)
-			nm = basename(nm);
+			nm = mybasename(nm);
 		memcpy(hdr->name, nm, MIN(NAMELEN, strlen(nm)));
 		struct stat stbf;
 		if (nm && 0 == stat(state->opts->iname, &stbf)) {
@@ -265,7 +254,7 @@ void lzo_hdr(header_t* hdr, loff_t hole, lzo_state *state)
 	state->hdr_seen = sizeof(header_t);
 }
 
-int lzo_parse_hdr(unsigned char* bf, loff_t* hole, lzo_state *state)
+int lzo_parse_hdr(unsigned char* bf, loff_t* hole, lzo_state *state, int eof)
 {
 	header_t *hdr = (header_t*)bf;
 	if (ntohs(hdr->version_needed_to_extract) > 0x1030 && ntohs(hdr->version_needed_to_extract) != F_VERSION) {
@@ -312,8 +301,8 @@ int lzo_parse_hdr(unsigned char* bf, loff_t* hole, lzo_state *state)
 	uint32_t comp = (state->flags & F_H_CRC32 ? lzo_crc32(  CRC32_INIT_VALUE, (const lzo_bytep)hdr, sizeof(header_t)-NAMELEN+hdr->nmlen-4)
 						: lzo_adler32(ADLER32_INIT_VALUE, (const lzo_bytep)hdr, sizeof(header_t)-NAMELEN+hdr->nmlen-4));
 	if (cksum != comp) {
-		FPLOG(FATAL, "header fails checksum %08x != %08x\n",
-			cksum, comp);
+		FPLOG(eof? WARN: FATAL, "header (offs %li) fails checksum %08x != %08x\n",
+			bf-state->obuf, cksum, comp);
 		return -6;
 	}
 	int off = sizeof(header_t) + hdr->nmlen-NAMELEN;
@@ -420,13 +409,15 @@ int lzo_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 		char* next = strchr(param, ':');
 		if (next)
 			*next++ = 0;
+		size_t length = strlen(param);
+
 		if (!strcmp(param, "help"))
 			FPLOG(INFO, "%s", lzo_help);
-		else if (!memcmp(param, "compr", 5))
+		else if (length >= 5 && !memcmp(param, "compr", 5))
 			state->mode = COMPRESS;
-		else if (!memcmp(param, "decom", 5))
+		else if (length >= 5 && !memcmp(param, "decom", 5))
 			state->mode = DECOMPRESS;
-		else if (!memcmp(param, "bench", 5))
+		else if (length >= 5 && !memcmp(param, "bench", 5))
 			state->do_bench = 1;
 		else if (!strcmp(param, "search"))
 			state->do_search = 1;
@@ -434,17 +425,17 @@ int lzo_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 			state->debug = 1;
 		else if (!strcmp(param, "crc32"))
 			state->flags = (state->flags | F_H_CRC32 | F_CRC32_C | F_CRC32_D) & ~(F_ADLER32_C | F_ADLER32_D);
-		else if (!memcmp(param, "opt", 3))
+		else if (length >= 3 && !memcmp(param, "opt", 3))
 			state->do_opt = 1;
-		else if (!memcmp(param, "nodisc", 6))
+		else if (length >= 6 && !memcmp(param, "nodisc", 6))
 			state->nodiscard = 1;
-		else if (!memcmp(param, "algo=", 5))
+		else if (length >= 5 && !memcmp(param, "algo=", 5))
 			err += choose_alg(param+5, state);
-		else if (!memcmp(param, "alg=", 4))
+		else if (length >= 4 && !memcmp(param, "alg=", 4))
 			err += choose_alg(param+4, state);
-		else if (!memcmp(param, "algorithm=", 10))
+		else if (length >= 10 && !memcmp(param, "algorithm=", 10))
 			err += choose_alg(param+10, state);
-		else if (!memcmp(param, "flags=", 6)) {
+		else if (length >= 6 && !memcmp(param, "flags=", 6)) {
 			state->flags = strtol(param+6, NULL, 0);
 			/* TODO Sanity check for flags ... */
 			//FPLOG(INFO, "Flags: %08x\n", state->flags);
@@ -459,62 +450,17 @@ int lzo_plug_init(void **stat, char* param, int seq, const opt_t *opt)
 	return err;
 }
 
-void* slackalloc(size_t ln, lzo_state *state)
-{
-	unsigned char *ptr = (unsigned char*)malloc(ln+state->slackpre+state->slackpost+pagesize);
-	if (!ptr) {
-		FPLOG(FATAL, "allocation of %i bytes failed: %s\n",
-			ln+state->slackpre+state->slackpost, strerror(errno));
-		raise(SIGQUIT);
-	}
-	state->orig_dbuf = ptr;
-	ptr += state->slackpre + pagesize-1;
-	ptr -= (unsigned long)ptr%pagesize;
-	return ptr;
-}
-
-void* slackrealloc(void* base, size_t newln, lzo_state *state)
-{
-	unsigned char *ptr, *optr;
-	++state->nr_realloc;
-	/* Note: We could use free and malloc IF we have no data decompressed yet 
-	 * (d_off == 0) and no slack space from plugins behind us is needed.
-	 * Probably not worth the effort ... */
-	ptr = (unsigned char*)malloc(newln+state->slackpre+state->slackpost+pagesize);
-	/* Note: We can be somewhat graceful if realloc fails by returning the original
-	 * pointer and buffer size and raise(SIGQUIT) -- this would result in 
-	 * writing out data that has been processed already.
-	 */
-	if (!ptr) {
-		FPLOG(FATAL, "reallocation of %i bytes failed: %s\n",
-			newln+state->slackpre+state->slackpost, strerror(errno));
-		raise(SIGQUIT);
-		return NULL;
-	}
-	optr = ptr;
-	ptr += state->slackpre + pagesize-1;
-	ptr -= (unsigned long)ptr%pagesize;
-	memcpy(ptr-state->slackpre, (char*)base-state->slackpre, state->dbuflen+state->slackpre+state->slackpost);
-	free(state->orig_dbuf);
-	state->orig_dbuf = optr;
-	return ptr;
-}
-
-void slackfree(void* base, lzo_state *state)
-{
-	//free(base-state->slackpre);
-	free(state->orig_dbuf);
-}
-
 int lzo_plug_release(void **stat)
 {
 	if (!stat | !*stat)
 		return -1;
 	lzo_state *state = (lzo_state*)*stat;
 	if (state->dbuflen)
-		slackfree(state->dbuf, state);
+		slack_free(state->dbuf, state->slackpre);
 	if (state->workspace)
 		free(state->workspace);
+	if (state->buf_zero)
+		slack_free(state->buf_zero, state->slackpre);
 	free(*stat);
 	return 0;
 }
@@ -523,7 +469,7 @@ int lzo_plug_release(void **stat)
 #define MAXBLOCKSZ 16UL*1024UL*1024UL
 int lzo_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 	     unsigned int totslack_pre, unsigned int totslack_post,
-	     const fstate_t *fst, void **stat)
+	     const fstate_t *fst, void **stat, int islast)
 {
 	lzo_state *state = (lzo_state*)*stat;
 	state->opts = opt;
@@ -557,9 +503,11 @@ int lzo_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 	} else {
 		state->dbuflen = 4*bsz+16;
 	}
-	state->slackpost = totslack_post;
 	state->slackpre  = totslack_pre ;
-	state->dbuf = (unsigned char*)slackalloc(state->dbuflen, state);
+	state->slackpost = totslack_post;
+	state->islast = islast;
+	state->dbuf = (unsigned char*)slack_alloc(state->dbuflen, totslack_pre, totslack_post,
+						  CLEAR_NONE, &ddr_plug, state->seq);
 	if (state->do_bench) 
 		state->cpu = 0;
 	if (state->mode == COMPRESS) {
@@ -570,7 +518,7 @@ int lzo_open(const opt_t *opt, int ilnchg, int olnchg, int ichg, int ochg,
 			FPLOG(WARN, "Blocks larger than 256kiB need recompilation of lzop (%ikiB specified)\n",
 				opt->softbs>>10);
 	}
-	state->next_ipos = opt->init_ipos;
+	state->next_ipos = state->mode == COMPRESS? opt->init_ipos: -1;
 	return 0;
 	/* This breaks MD5 in chain before us
 	return consumed;
@@ -638,6 +586,8 @@ int encode_hole(unsigned char* bhdp, int nopre, loff_t hsz, int hlen, lzo_state
 		return encode_hole_swap(bhdp, nopre, hsz, hlen, state);
 	hlen = sizeof(lzop_hdr)+sizeof(header_t)+4;
 	unsigned char* ptr = bhdp - (nopre? 0: hlen);
+	LZO_DEBUG(FPLOG(DEBUG, " hole ptr %p, zero %p, obuf %p, dbuf %p\n",
+			ptr, state->buf_zero, state->obuf, state->dbuf));
 	memset(ptr, 0, 4);	/* EOF */
 	ptr += 4;
 	memcpy(ptr, lzop_hdr, sizeof(lzop_hdr));
@@ -651,12 +601,14 @@ unsigned char* lzo_compress(fstate_t *fst, unsigned char *bf,
 			    int *towr, int eof, int *recall, lzo_state *state)
 {
 	//const loff_t ooff = fst->opos;
-	lzo_uint dst_len = state->dbuflen-3-sizeof(lzop_hdr)-sizeof(header_t);
-	unsigned char *hdrp = state->dbuf+3+sizeof(lzop_hdr);
+	lzo_uint dst_len = state->dbuflen-4-sizeof(lzop_hdr)-sizeof(header_t);
+	unsigned char *hdrp = state->dbuf+4+sizeof(lzop_hdr);
 	unsigned char *bhdp = hdrp+sizeof(header_t);
 	unsigned char *wrbf = bhdp;
 	unsigned int addwr = 0;
 	unsigned int hlen = sizeof(blockhdr_t)-4+((state->flags&(F_ADLER32_C|F_CRC32_C))? 4: 0);
+	FPLOG(DEBUG, "lzo_compress ipos %li expected %li block #%i towr %i eof %i\n",
+		fst->ipos, state->next_ipos, state->blockno, *towr, eof);
 	if (state->hdr_seen == 0) { // was: ooff == state->opts->init_opos) {
 		if (state->opts->init_opos > 0 && state->opts->extend) {
 			/* TODO: Check for multipart archive and attach a part if yes */
@@ -669,7 +621,7 @@ unsigned char* lzo_compress(fstate_t *fst, unsigned char *bf,
 				FPLOG(FATAL, "Can only extend lzo files with existing magic\n", ln);
 				abort();
 			};
-			if (lzo_parse_hdr(bhdp+sizeof(lzop_hdr), NULL, state) < 0)
+			if (lzo_parse_hdr(bhdp+sizeof(lzop_hdr), NULL, state, 0) < 0)
 				abort();
 			/* TODO (optional): Jump block headers to see whether we are at a valid offset */
 			/* Due to different flags, blkhdr size could have changed ... */
@@ -686,21 +638,21 @@ unsigned char* lzo_compress(fstate_t *fst, unsigned char *bf,
 			}
 		}
 		if (state->hdr_seen == 0) {
-			memcpy(state->dbuf+3, lzop_hdr, sizeof(lzop_hdr));
+			memcpy(state->dbuf+4, lzop_hdr, sizeof(lzop_hdr));
 			lzo_hdr((header_t*)hdrp, 0, state);
 			addwr = sizeof(header_t) + sizeof(lzop_hdr);
-			wrbf = state->dbuf+3;
+			wrbf = state->dbuf+4;
 			state->cmp_hdr += sizeof(lzop_hdr)+sizeof(header_t);
 		}
 	}
 	if (fst->ipos > state->next_ipos) {
 		/* Sparse support */
 		const loff_t hsz = fst->ipos - state->next_ipos;
-		if (state->debug)
-			FPLOG(DEBUG, "hole %i@%i/%i (sz %i/%i+%i)\n",
-				state->blockno, state->next_ipos, fst->opos-hsz,
-				hsz, 0, hlen);
 		int holehdrsz = encode_hole(bhdp, addwr, hsz, hlen, state);
+		if (state->debug)
+			FPLOG(DEBUG, "hole %i@%i/%i (sz %i/%i+0)\n",
+				state->blockno, state->next_ipos, fst->opos,
+				hsz, holehdrsz);
 		if (!addwr)
 			wrbf -= holehdrsz;
 		else 
@@ -709,8 +661,6 @@ unsigned char* lzo_compress(fstate_t *fst, unsigned char *bf,
 		addwr += holehdrsz;
 		state->next_ipos = fst->ipos;
 		state->blockno++;
-		/* Compensate for dd_rescue moving opos forward ... */
-		fst->opos -= hsz;
 	}
 	/* NOTE: We always calc checksum of uncompressed data, as we don't get a
 	 * checksum at all otherwise (lzop decompressor does not allow for checksums
@@ -812,17 +762,33 @@ unsigned char* lzo_search_hdr(fstate_t *fst, unsigned char* bf, int *towr,
 		if (unc_len == lzo1 && cmp_len == lzo2 
 				&& fst->buf[off+8] == lzop_hdr[8]) {
 			loff_t hole;
-			int hlen = lzo_parse_hdr(fst->buf+off+sizeof(lzop_hdr), &hole, state);
-			if (!state->opts->quiet)
-				FPLOG(INFO, "lzop header at %i (sz %i/hole %li)\n", fst->ipos+off, 
-					hlen+sizeof(lzop_hdr), hole);
-			fst->opos += hole;
-			off += hlen+sizeof(lzop_hdr);
-			unc_len = *(uint32_t*)(fst->buf+off);
-			cmp_len = *(uint32_t*)(fst->buf+off+4);
-			if (state->debug)
-				FPLOG(DEBUG, "Next blk: %i/%i\n",
-					ntohl(unc_len), ntohl(cmp_len));
+			int hlen = lzo_parse_hdr(fst->buf+off+sizeof(lzop_hdr), &hole, state, eof);
+			FPLOG(INFO, "lzop header at %i (sz %i+%i/hole %li)\n", fst->ipos+off,
+				hlen, sizeof(lzop_hdr), hole);
+			/* FIXME: This optimization, just jumping over the hole,
+			 * is only valid if there is no plugin behind us in the chain.
+			 * Otherwise, we need to send down zeroes, just to have them
+			 * eliminated again down the line in case -a is used (and in case
+			 * they are not transformed to non-zeroes again)a
+			 */
+			if (state->islast) {
+				fst->opos += hole;
+				off += hlen+sizeof(lzop_hdr);
+				unc_len = *(uint32_t*)(fst->buf+off);
+				cmp_len = *(uint32_t*)(fst->buf+off+4);
+				if (state->debug)
+					FPLOG(DEBUG, "Next blk: %i/%i\n",
+						ntohl(unc_len), ntohl(cmp_len));
+			} else {
+				FPLOG(WARN, "not yet implmented: return hole @opos %li\n",
+					fst->opos);
+				fst->opos += hole;
+				off += hlen+sizeof(lzop_hdr);
+				unc_len = *(uint32_t*)(fst->buf+off);
+				cmp_len = *(uint32_t*)(fst->buf+off+4);
+				//*recall = RECALL_MARK;
+				//*towr = hole;
+			}
 		}
 		if (unc_len & mask || cmp_len & mask)
 			continue;
@@ -888,9 +854,8 @@ unsigned char* lzo_search_hdr(fstate_t *fst, unsigned char* bf, int *towr,
 					}
 				}
 			}
-			if (!state->opts->quiet)
-				FPLOG(INFO, "Found block @ %i (flags %08x)\n",
-					fst->ipos+off, state->flags);
+			FPLOG(INFO, "Found block @ %i (flags %08x)\n",
+				fst->ipos+off, state->flags);
 			//*towr -= off;
 			state->hdroff = off;
 			state->do_search = 0;
@@ -901,8 +866,7 @@ unsigned char* lzo_search_hdr(fstate_t *fst, unsigned char* bf, int *towr,
 			const size_t totbufln = state->opts->softbs - ddr_plug.slack_post*((state->opts->softbs+15)/16);
 			const size_t left = totbufln - (*towr-off);
 			if (left < state->opts->softbs) {
-				if (!state->opts->quiet)
-					FPLOG(INFO, "Buffer exhausted Blk Cand @ %i\n", fst->ipos+off);
+				FPLOG(INFO, "Buffer exhausted Blk Cand @ %i\n", fst->ipos+off);
 				off += fst->buf-state->obuf;
 				fst->buf = state->obuf;
 				assert(off >= 0);
@@ -913,6 +877,7 @@ unsigned char* lzo_search_hdr(fstate_t *fst, unsigned char* bf, int *towr,
 					fst->ipos+off, off, fst->buf, state->obuf);
 			if (state->obuf != fst->buf+off)
 				memmove(state->obuf, fst->buf+off, *towr-off);
+			/* Move buffer pointer, compensate with hdroff */
 			fst->buf = state->obuf + *towr-off;
 			state->hdroff = -(*towr-off);
 			*towr = 0;
@@ -964,6 +929,9 @@ int recover_decompr_error(lzo_state *state, fstate_t *fst,
 	if (recoverable && !state->nodiscard) {
 		state->cmp_hdr += bhsz;
 		*c_off += cmp_len+bhsz;
+		// Note: This only works if we're last in line
+		if (!state->islast)
+			FPLOG(WARN, "r_d_e: Skipping while not last in line\n");
 		//Don't d_off += dst_len, as we're skipping; instead:
 		fst->opos += unc_len;
 		state->cmp_ln += cmp_len;
@@ -977,11 +945,31 @@ int recover_decompr_error(lzo_state *state, fstate_t *fst,
 
 #define QUIT { raise(SIGQUIT); ++do_break; break; }
 #define BREAK if (!state->nodiscard) ++do_break; break
-#define DRAIN(x) { do { ++do_break; *recall=1; 		\
-		   LZO_DEBUG(FPLOG(DEBUG, "Drain %i bytes before %s error handling\n", d_off, x));	\
-		   eof = 0;				\
-       		   break; } while(0); 			\
+#define DRAIN(x) { do { ++do_break; *recall = RECALL_MARK;	\
+		   FPLOG(DEBUG, "Drain %i bytes before %s error handling\n", d_off, x);	\
+		   eof = 0;					\
+		   break; } while(0); 				\
 		   if (do_break) break; }
+#define DRAINH(x) { do { ++do_break; 				\
+		   FPLOG(DEBUG, "Drain %i bytes before %s handling\n", d_off, x);	\
+		   eof = 0;					\
+		   break; } while(0); 				\
+		   if (do_break) break; }
+
+
+/* Output zero-filled blocks on holes if we're not the last in line */
+unsigned char* lzo_decompress_hole(fstate_t *fst, int *towr, lzo_state *state)
+{
+	const int ln = MIN(state->inhole, state->opts->softbs);
+	if (!state->buf_zero)
+       		state->buf_zero = slack_alloc(state->opts->softbs, state->slackpre, state->slackpost,
+					      CLEAR_ALL, &ddr_plug, state->seq);
+	FPLOG(DEBUG, "zero out hole (left %i, process %i)\n",
+		state->inhole, ln);
+	state->inhole -= ln;
+	*towr = ln;
+	return state->buf_zero;
+}
 
 /* TODO:
  * - Debug: Output block boundaries
@@ -992,10 +980,25 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 			      int eof, int *recall, lzo_state *state)
 {
 	const loff_t ooff = fst->opos;
-	const int inlen = *towr;
+	/* FIXME: How do we avoid processing things twice? When someone
+	 * (not necessarily ourselves, but possibly also ahead of us in the chain)
+	 * has returned RECALL_MARK, we will get the same buffer wuth same ipos
+	 * presented again. We need to detect this.
+	 * We we inserted a zeroed hole ourselves, we will have saved_c_off,
+	 * which seems enough to handle this. How about other cases with
+	 * RECALL_MARK?
+	 */
+	const int inlen = (fst->ipos == state->next_ipos && !state->saved_c_off)? 0: *towr;
+	//const int inlen = *towr;
 	/* Decompression is tricky */
-	int c_off = 0;
+	int c_off = state->saved_c_off;
 	int d_off = 0;
+	if (state->inhole) {
+		*recall = RECALL_MARK;
+		return lzo_decompress_hole(fst, towr, state);
+	} else
+		state->saved_c_off = 0;
+	state->next_ipos = fst->ipos;
 	if (!state->hdr_seen) {
 		assert(ooff - state->opts->init_opos == 0);
 		if (memcmp(bf, lzop_hdr, sizeof(lzop_hdr))) {
@@ -1012,13 +1015,13 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 					FPLOG(FATAL, "lzop magic broken\n");
 					abort();
 				}
-				if (lzo_parse_hdr(state->dbuf+sizeof(lzop_hdr), NULL, state) < 0)
+				if (lzo_parse_hdr(state->dbuf+sizeof(lzop_hdr), NULL, state, 0) < 0)
 					abort();
 			}
-		} else {	
+		} else {
 			state->cmp_hdr = sizeof(lzop_hdr);
 			c_off += sizeof(lzop_hdr);
-			int err = lzo_parse_hdr(bf+c_off, NULL, state);
+			int err = lzo_parse_hdr(bf+c_off, NULL, state, 0);
 			if (err < 0)
 				abort();
 			c_off += err;
@@ -1027,9 +1030,14 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 	/* Now do processing: Do we have a full block? */
 	const size_t totbufln = state->opts->softbs - ddr_plug.slack_post*((state->opts->softbs+15)/16);
 	unsigned char* effbf = NULL;
-	size_t have_len = 0;
+	ssize_t have_len = 0;
 	uint32_t cmp_len = 0, unc_len = 0;
 	int bhsz = sizeof(blockhdr_t);
+	FPLOG(DEBUG, "decompress ipos %" LL "d (next %" LL "i)/opos %" LL "d, inlen %i, hdroff %i\n",
+		fst->ipos, state->next_ipos, fst->opos, inlen, state->hdroff);
+	/* Ensure hdroff is within range */
+	assert(fst->buf + state->hdroff >= state->obuf - 7);
+	/* ??? */
 	if (inlen-state->hdroff <= 0)
 		return bf;
 	/* Main loop: Process blocks */
@@ -1038,13 +1046,16 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 		char is_err = 0;
 		effbf = bf+c_off+state->hdroff;
 		lzo_uint dst_len;
-		LZO_DEBUG(FPLOG(DEBUG, "dec blk @ %p (offs %i, stoffs %i, bln %zi, tbw %i)\n",
-				effbf, effbf-state->obuf, state->hdroff, totbufln, inlen));
 		blockhdr_t *hdr = (blockhdr_t*)effbf;
 		have_len = inlen-state->hdroff-c_off;
+		LZO_DEBUG(FPLOG(DEBUG, " dec blk @ %p (offs %i, stoffs %i, bln %zi, tbw %i have %i)\n",
+				effbf, effbf-state->obuf, state->hdroff, totbufln, inlen, have_len));
 		/* Not even space for unc_len/EOF */
-		if (have_len < 4)
+		if (have_len < 4) {
+			unc_len = 0;
+			bhsz = 0;
 			break;
+		}
 		unc_len = ntohl(hdr->uncmpr_len);
 		bhsz = bhdr_size(state, unc_len, cmp_len);
 		if (!unc_len && (!(state->flags & F_MULTIPART) || (eof && have_len < 8))) {
@@ -1056,16 +1067,26 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 					state->cmp_ln+state->cmp_hdr);
 			break;
 		}
-		if (!unc_len && state->flags & F_MULTIPART && have_len > 32) {
+		if (!unc_len && state->flags & F_MULTIPART && have_len >= sizeof(lzop_hdr)+sizeof(header_t)+4) {
 			/* EOF with new LZOP sig */
-			LZO_DEBUG(FPLOG(DEBUG, "Next part ...\n"));
+			LZO_DEBUG(FPLOG(DEBUG, " Next part ...\n"));
 			if (memcmp(effbf+4, lzop_hdr, sizeof(lzop_hdr))) {
-				FPLOG(FATAL, "EOF with MULTIPART, but no new hdr\n");
-				raise(SIGQUIT);
+				FPLOG(eof? WARN: FATAL, "EOF with MULTIPART, but no new hdr (%zi bytes)\n",
+					have_len);
+				//DRAIN("EOF");
+				//raise(SIGQUIT);
+				fst->nrerr++;
 				break;
 			}
 			loff_t hsz;
-			int hln = lzo_parse_hdr(effbf+4+sizeof(lzop_hdr), &hsz, state);
+			int hln = lzo_parse_hdr(effbf+4+sizeof(lzop_hdr), &hsz, state, eof);
+			if (hln < 0) {
+				FPLOG(eof? INFO: WARN, "faulty header @ %li: %i, have %i/%i eof %i\n",
+					/*fst->ipos+*/(effbf-state->obuf+4+sizeof(lzop_hdr)), hln, have_len,
+					sizeof(lzop_hdr)+sizeof(header_t)+4, eof);
+				if (!eof)
+					fst->nrerr++;
+			}
 			bhsz = hln+sizeof(lzop_hdr)+4;
 			if (!hsz) {
 				if (!state->opts->quiet)
@@ -1098,7 +1119,7 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 		if (cmp_len == unc_len) 
 			cmp_cksum = unc_cksum;
 		
-		LZO_DEBUG(FPLOG(DEBUG, "dec blk @ %p (hdroff %i, cln %i, uln %i, have %i)\n",
+		LZO_DEBUG(FPLOG(DEBUG, " dec blk @ %p (hdroff %i, cln %i, uln %i, have %i)\n",
 				effbf, c_off+state->hdroff, unc_len, cmp_len, have_len));
 		/* Block incomplete? Then we're done for this round ... */
 		if (have_len < bhsz+cmp_len)
@@ -1129,19 +1150,27 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 
 		/* Sparse ... */
 		if (0 == cmp_len) {
-			if (d_off) 
-				DRAIN("hole");
+			if (d_off)
+				DRAINH("hole");
 			if (state->debug)
-				FPLOG(DEBUG, "hole %i@%i/%i (sz %i+%i/%i)\n",
+				FPLOG(DEBUG, "hole %i@%i+%i/%i (sz %i+%i/%i)\n",
 					state->blockno, fst->ipos+c_off+state->hdroff,
-					fst->opos+d_off, bhsz, cmp_len, unc_len);
+					inlen, fst->opos+d_off, bhsz, cmp_len, unc_len);
 			state->cmp_hdr += bhsz;
 			c_off += cmp_len+bhsz;
 			//Instead of d_off += unc_len;
-			fst->opos += unc_len;
 			state->cmp_ln += cmp_len;
 			//state->unc_ln += unc_len;
 			state->blockno++;
+			if (!state->islast) {
+				state->inhole = unc_len;
+				*recall = RECALL_MARK;
+				/* Set up state to continue cleanly after the hole */
+				state->saved_c_off = c_off;
+				/* Alternatively, we could have moved fst->buf and hdroff ? */
+				return lzo_decompress_hole(fst, towr, state);
+			} else
+				fst->opos += unc_len;
 			continue;
 		}
 
@@ -1155,12 +1184,12 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 						unc_len, state->cmp_ln+state->cmp_hdr);
 			size_t newlen = unc_len+d_off+255;
 			newlen -= newlen%256;
-			unsigned char *newbuf = (unsigned char*)slackrealloc(state->dbuf, newlen, state);
+			state->dbuf = slack_realloc(state->dbuf, newlen, state->slackpre, state->slackpost,
+						    CLEAR_NONE, &ddr_plug, state->seq);
 			/* if realloc failed, exit loop, write out existing data and exit;
 			 * slackrealloc has done raise(SIGQUIT) already ... */
-			if (!newbuf)
+			if (!state->dbuf)
 				break;
-			state->dbuf = newbuf;
 			state->dbuflen = newlen;
 			dst_len = newlen-d_off;
 		}
@@ -1175,7 +1204,7 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 				FPLOG(WARN, "compressed %i > uncompressed %i breaks lzop\n",
 					cmp_len, unc_len);
 			err = state->algo->decompr(effbf+bhsz, cmp_len, (unsigned char*)state->dbuf+d_off, &dst_len, NULL);
-			LZO_DEBUG(FPLOG(DEBUG, "decompressed %i@%p -> %i\n",
+			LZO_DEBUG(FPLOG(DEBUG, " decompressed %i@%p -> %i\n",
 				cmp_len, effbf+bhsz, dst_len));
 			if (dst_len != unc_len) {
 				fst->nrerr++;
@@ -1294,7 +1323,10 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 	/* Need drain? */
 	int spcleft = totbufln-(fst->buf+inlen-state->obuf);
 	//if (!*recall && spcleft < state->opts->softbs)
-	//       *recall = 1;
+	//       *recall = RECALL_MARK;
+	/* FIXME */
+	//assert(have_len >= 0);
+	//int nextrd = *recall != RECALL_NONE? 0: state->opts->softbs;
 	int nextrd = *recall? 0: state->opts->softbs;
 	/* Trivial case: No bytes to store */
 	if (have_len == 0) {
@@ -1309,11 +1341,11 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 		state->hdroff = -have_len;
 		fst->buf = state->obuf+have_len;
 	/* Enough space for complete block and enough left for next read? */
-	} else if (effbf+bhsz+cmp_len <= state->obuf+totbufln && spcleft >= nextrd) {
+	} else if ((effbf+bhsz+cmp_len <= state->obuf+totbufln && spcleft >= nextrd) || *recall == RECALL_MARK) {
 		/* We have enough space to just append */
 		state->hdroff -= inlen-c_off;
 		fst->buf += inlen;
-		LZO_DEBUG(FPLOG(DEBUG, "append  @ %p\n", fst->buf));
+		LZO_DEBUG(FPLOG(DEBUG, " append  @ %p\n", fst->buf));
 
 	/* OK, now for the bad cases:
 	 * (a) We can't append, but everything fits if we memmove to start
@@ -1321,9 +1353,9 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 	 */
 	/* Simplify to bhsz+cmp_len+state->softbs < totbufln ? */
 	} else if (bhsz+cmp_len < totbufln &&
-		   have_len+nextrd < totbufln) {
+		   have_len+nextrd < totbufln && have_len >= 0) {
 		/* We need to move block to beg of buffer */
-		LZO_DEBUG(FPLOG(DEBUG, "move %i bytes to buffer head\n", have_len));
+		LZO_DEBUG(FPLOG(DEBUG, " move %i bytes to buffer head\n", have_len));
 		if (effbf != state->obuf) {
 			memmove(state->obuf, effbf, have_len);
 			++state->nr_memmove;
@@ -1341,10 +1373,13 @@ unsigned char* lzo_decompress(fstate_t *fst, unsigned char* bf, int *towr,
 	}
 
 	*towr = d_off;
+	//if (*recall != RECALL_NONE)
+	//	state->saved_c_off = c_off;
 	return state->dbuf;
 }
 #undef BREAK
 #undef DRAIN
+#undef DRAINH
 
 
 unsigned char* lzo_block(fstate_t *fst, unsigned char* bf, 
@@ -1358,7 +1393,7 @@ unsigned char* lzo_block(fstate_t *fst, unsigned char* bf,
 	if (state->do_bench) 
 		t1 = clock();
 	if (state->mode == COMPRESS) 
-		ptr = lzo_compress(  fst, bf, towr, eof, recall, state);
+		ptr = lzo_compress(fst, bf, towr, eof, recall, state);
 	else {
 		if (state->do_search) 
 			ptr = lzo_search_hdr(fst, bf, towr, eof, recall, state);
@@ -1394,8 +1429,8 @@ int lzo_close(loff_t ooff, void **stat)
 					state->nr_realloc, state->dbuflen/1024,
 					state->nr_memmove, state->nr_cheapmemmove);
 		}
-		/* Only output if it took us more than 0.05s, otherwise it's completely meaningless */
-		if (state->do_bench && state->cpu/(CLOCKS_PER_SEC/20) > 0)
+		/* Only output if it took us more than 0.01s, otherwise it's completely meaningless */
+		if (state->do_bench && state->cpu/(CLOCKS_PER_SEC/100) > 0)
 			FPLOG(INFO, "%.2fs CPU time, %.1fMiB/s\n",
 				(double)state->cpu/CLOCKS_PER_SEC, 
 				state->unc_ln/1024 / (state->cpu/(CLOCKS_PER_SEC/1024.0)));
